--------------------------- MODULE Filters ---------------------------
EXTENDS SequenceTheorems, Types

\* Will go into module SequenceTheorems, left here until standard library is updated
LEMMA mySubSeqOfSubSeq ==
  ASSUME NEW S, NEW s \in Seq(S),
         NEW m \in 1 .. Len(s)+1,
         NEW n \in m-1 .. Len(s),
         NEW i \in 1 .. n-m+2,
         NEW j \in i-1 .. n-m+1
  PROVE  SubSeq( SubSeq(s,m,n), i, j ) = SubSeq(s, m+i-1, m+j-1)
<1>. DEFINE smn == SubSeq(s,m,n)
            lhs == SubSeq(smn, i, j)
            rhs == SubSeq(s, m+i-1, m+j-1)
<1>. SUFFICES lhs = rhs
  OBVIOUS
<1>1. /\ smn \in Seq(S)
      /\ Len(smn) = n-m+1
      /\ \A k \in 1 .. n-m+1 : smn[k] = s[m+k-1]
  BY SubSeqProperties, Zenon
<1>. HIDE DEF smn
<1>2. /\ lhs \in Seq(S)
      /\ Len(lhs) = j-i+1
      /\ \A k \in 1 .. j-i+1 : lhs[k] = smn[i+k-1]
  BY <1>1, i \in 1 .. Len(smn)+1, j \in i-1 .. Len(smn), SubSeqProperties, Zenon
<1>3. /\ rhs \in Seq(S)
      /\ Len(rhs) = (m+j-1) - (m+i-1) + 1
      /\ \A k \in 1 .. (m+j-1) - (m+i-1) + 1 : rhs[k] = s[(m+i-1)+k-1]
  BY m+i-1 \in 1 .. Len(s)+1, m+j-1 \in (m+i-1)-1 .. Len(s), SubSeqProperties, Zenon
<1>. HIDE DEF lhs, rhs
<1>4. Len(lhs) = Len(rhs)
  BY <1>2, <1>3
<1>5. \A k \in 1 .. Len(lhs) : lhs[k] = rhs[k]
  BY <1>1, <1>2, <1>3
<1>. QED  BY <1>2, <1>3, <1>4, <1>5, SeqEqual

LEMMA myRangeIsPrefix == 
  ASSUME NEW S, NEW s \in Seq(S), NEW t \in Seq(S),
         IsPrefix(s,t)
  PROVE Range(s) \subseteq Range(t)
BY IsPrefixProperties, RangeConcatenation

LEMMA myIsPrefixMap ==
  ASSUME NEW S, NEW s \in Seq(S), NEW t \in Seq(S), IsPrefix(s,t),
         NEW Op(_)
  PROVE  IsPrefix([i \in 1 .. Len(s) |-> Op(s[i])],
                  [i \in 1 .. Len(t) |-> Op(t[i])])
<1>. DEFINE MapS == [i \in 1 .. Len(s) |-> Op(s[i])]
            MapT == [i \in 1 .. Len(t) |-> Op(t[i])]
            U == { Op(x) : x \in S }
<1>1. /\ Len(s) <= Len(t)
      /\ s = SubSeq(t, 1, Len(s))
  BY IsPrefixProperties
<1>2. /\ \A i \in 1 .. Len(s) : Op(s[i]) \in U
      /\ \A i \in 1 .. Len(t) : Op(t[i]) \in U
  OBVIOUS
<1>3. /\ MapS \in Seq(U)
      /\ MapT \in Seq(U)
  BY <1>2
<1>4. /\ Len(MapS) = Len(s)
      /\ Len(MapT) = Len(t)
  OBVIOUS
<1>5. MapS = SubSeq(MapT, 1, Len(MapS))
  BY <1>1
<1>. QED  BY <1>1, <1>3, <1>4, <1>5, IsPrefixProperties, Zenon

(***************************************************************************)
(* Project a sequence of system states to a sequence of states of an agent.*)
(***************************************************************************)
project_agent(H,a) == [i \in 1 .. Len(H) |-> H[i].st[a]]

LEMMA project_agent_type ==
  ASSUME NEW H \in Seq(SystemState), NEW a \in Agent
  PROVE  /\ project_agent(H,a) \in Seq(State)
         /\ Len(project_agent(H,a)) = Len(H)
BY DEF project_agent, SystemState

LEMMA project_agent_append ==
  ASSUME NEW H \in Seq(SystemState), NEW syst \in SystemState,
         NEW a \in Agent
  PROVE  project_agent(Append(H,syst), a) = Append(project_agent(H,a), syst.st[a])
BY DEF project_agent

LEMMA project_agent_last ==
  ASSUME NEW H \in Seq(SystemState), H # << >>, NEW a \in Agent
  PROVE  Last(project_agent(H,a)) = Last(H).st[a]
BY DEF project_agent, Last

(***************************************************************************)
(* Remove stuttering from a finite sequence.                               *)
(***************************************************************************)
unstutter_body(H,f) ==
    IF H = << >> THEN H \* Important for the proof of unstutter_type
    ELSE IF Len(H) \leq 1 THEN H
    ELSE IF Last(H) = H[Len(H)-1] THEN f[Front(H)]
    ELSE Append(f[Front(H)], Last(H))
    
unstutter == 
    LET f[H \in Seq(State)] == unstutter_body(H,f)
    IN f


LEMMA unstutter_type ==
  /\ unstutter = [ H \in Seq(State) |-> unstutter_body(H,unstutter) ]
  /\ unstutter \in [Seq(State) -> Seq(State)]
<1> DEFINE fdef(unstuttered, H) == 
         IF Len(H) \leq 1 THEN H
         ELSE IF Last(H) = H[Len(H)-1] THEN unstuttered
         ELSE Append(unstuttered, Last(H))  
<1> FrontInductiveDefHypothesis(unstutter, State, <<>>, fdef)
    BY DEF unstutter, unstutter_body, FrontInductiveDefHypothesis
<1> FrontInductiveDefConclusion(unstutter, State, <<>>, fdef)
    BY FrontInductiveDef, Zenon
<1> unstutter \in [Seq(State) -> Seq(State)]  
    <2> << >> \in Seq(State)
        BY Isa
    <2> ASSUME NEW v \in Seq(State), NEW H \in Seq(State), H # << >>
        PROVE fdef(v,H) \in Seq(State)
        BY LastProperties, Last(H) \in State, AppendProperties
    <2> QED
    BY FrontInductiveDefType, Isa
<1> QED
BY DEF FrontInductiveDefConclusion, unstutter_body

LEMMA unstutter_append ==
  ASSUME NEW H \in Seq(State), NEW s \in State
  PROVE  unstutter[Append(H,s)]
         = IF H = << >> \/ Last(H) # s THEN Append(unstutter[H], s)
           ELSE unstutter[H]
<1>. /\ Append(H,s) # << >>
     /\ Append(H,s) \in Seq(State)
     /\ Front(Append(H,s)) = H
     /\ Last(Append(H,s)) = s
  BY FrontLastAppend
<1>1. CASE H = << >>
  <2>. unstutter[H] = H
    BY <1>1, unstutter_type, Zenon DEF unstutter_body
  <2>. Len(Append(H,s)) <= 1
    BY <1>1
  <2>. QED  BY <1>1, unstutter_type, Zenon DEF unstutter_body
<1>2. CASE H # << >> /\ Last(H) = s
  <2>. /\ ~(Len(Append(H,s)) <= 1)
       /\ Last(Append(H,s)) = Append(H,s)[Len(Append(H,s))-1]
    BY <1>2 DEF Last
  <2>. QED  BY <1>2, unstutter_type, Zenon DEF unstutter_body
<1>3. CASE H # << >> /\ Last(H) # s
  <2>. /\ ~(Len(Append(H,s)) <= 1)
       /\ Last(Append(H,s)) # Append(H,s)[Len(Append(H,s))-1]
    BY <1>3 DEF Last
  <2>. QED  BY <1>3, unstutter_type, Zenon DEF unstutter_body
<1>. QED  BY <1>1, <1>2, <1>3

LEMMA unstutter_range ==
  \A H \in Seq(State) : Range(unstutter[H]) = Range(H)
<1>. DEFINE P(H) == Range(unstutter[H]) = Range(H)
<1>1. P(<<>>)
  <2>. << >> \in Seq(State)
    BY Isa
  <2>. QED
    BY unstutter_type, Zenon DEF unstutter_body
<1>2. ASSUME NEW H \in Seq(State), NEW s \in State, P(H)
      PROVE  P(Append(H,s))
  <2>. unstutter[H] \in Seq(State)
    BY unstutter_type
  <2>1. CASE H = << >> \/ Last(H) # s
    <3>. unstutter[Append(H,s)] = Append(unstutter[H], s)
      BY <2>1, unstutter_append
    <3>. /\ Range(unstutter[Append(H,s)]) = Range(unstutter[H]) \cup Range(<<s>>)
         /\ Range(Append(H,s)) = Range(H) \cup Range(<<s>>)
      BY AppendIsConcat, <<s>> \in Seq(State), RangeConcatenation, Isa
    <3>. QED  BY <1>2 
  <2>2. CASE H # << >> /\ Last(H) = s
    <3>. unstutter[Append(H,s)] = unstutter[H]
      BY <2>2, unstutter_append
    <3>. Range(Append(H,s)) = Range(H) \cup Range(<<s>>)
      BY AppendIsConcat, <<s>> \in Seq(State), RangeConcatenation, Isa
    <3>. Range(<<s>>) = {s}
      BY DEF Range
    <3>. s \in Range(H)
      BY <2>2, RangeEquality DEF Last
    <3>. QED  BY <1>2
  <2>. QED  BY <2>1, <2>2
<1>. QED  BY <1>1, <1>2, SequencesInductionAppend, Isa

LEMMA unstutter_empty ==
  ASSUME NEW H \in Seq(State)
  PROVE  unstutter[H] = <<>> <=> H = <<>>
<1>1. ASSUME H = << >>
      PROVE unstutter[H] = << >>
  BY <1>1, unstutter_type, Zenon DEF unstutter_body
<1>2. ASSUME H # << >>, unstutter[H] = << >> 
      PROVE  FALSE
  <2>1. Range(H) # {}
    BY <1>2, 1 \in 1 .. Len(H), RangeEquality, Isa
  <2>2. Range(unstutter[H]) = {}
    BY <1>2, unstutter_type, RangeEquality
  <2>. QED  BY <2>1, <2>2, unstutter_range
<1>. QED  BY <1>1, <1>2

LEMMA unstutter_first ==
  ASSUME NEW H \in Seq(State), H # << >>
  PROVE  unstutter[H][1] = H[1]
<1>. DEFINE P(S) == S # <<>> => unstutter[S][1] = S[1]
<1>1. P(<<>>)
  OBVIOUS
<1>2. ASSUME NEW S \in Seq(State), NEW s \in State, P(S)
      PROVE  P(Append(S,s))
  BY <1>2, unstutter_append, unstutter_empty
<1>3. \A S \in Seq(State) : P(S)
  BY <1>1, <1>2, SequencesInductionAppend, Isa
<1>. QED  BY <1>3

LEMMA unstutter_last ==
  ASSUME NEW H \in Seq(State), H # << >>
  PROVE  Last(unstutter[H]) = Last(H)
<1>. DEFINE P(S) == S # <<>> => Last(unstutter[S]) = Last(S)
<1>1. P(<<>>)
  OBVIOUS
<1>2. ASSUME NEW S \in Seq(State), NEW s \in State, P(S)
      PROVE  P(Append(S,s))
  BY <1>2, unstutter_append, unstutter_empty DEF Last
<1>3. \A S \in Seq(State) : P(S)
  BY <1>1, <1>2, SequencesInductionAppend, Isa
<1>. QED  BY <1>3

LEMMA unstutter_concat_stuttered_last ==
  ASSUME NEW S \in Seq(State), S # << >>, NEW T \in Seq(State),
         \A i \in 1 .. Len(T) : T[i] = Last(S)
  PROVE  unstutter[S \o T] = unstutter[S]
<1>. DEFINE P(U) == (\A i \in 1 .. Len(U) : U[i] = Last(S))
                    => unstutter[S \o U] = unstutter[S]
<1>1. P(<< >>)
  OBVIOUS
<1>2. ASSUME NEW U \in Seq(State), P(U), NEW s \in State
      PROVE  P(Append(U,s))
  <2>1. HAVE \A i \in 1 .. Len(Append(U,s)) : Append(U,s)[i] = Last(S)
  <2>2. /\ \A i \in 1 .. Len(U) : U[i] = Last(S)
        /\ s = Last(S)
    BY <2>1, AppendProperties DEF Last
  <2>3. S \o Append(U,s) = Append(S \o U, Last(S \o U))
    BY <2>2 DEF Last
  <2>4. unstutter[S \o Append(U,s)] = unstutter[S \o U]
    BY <2>3, unstutter_append DEF Last
  <2>. QED  BY <1>2, <2>2, <2>4
<1>3. \A U \in Seq(State) : P(U)
  BY <1>1, <1>2, SequencesInductionAppend, Isa
<1>. QED  BY <1>3

LEMMA unstutter_concat_different ==
  ASSUME NEW S \in Seq(State), NEW T \in Seq(State),
         S # << >> /\ T # << >> => T[1] # Last(S)
  PROVE  unstutter[S \o T] = unstutter[S] \o unstutter[T]
<1>. /\ unstutter[S] \in Seq(State)
     /\ unstutter[T] \in Seq(State)
     /\ << >> \in Seq(State)
     /\ unstutter[<< >>] \in Seq(State)
     /\ unstutter[<< >>] = << >>
  BY unstutter_type, << >> \in Seq(State), unstutter_empty, Zenon
<1>1. CASE S = << >>
  BY <1>1
<1>2. CASE S # << >>
  <2>. DEFINE P(U) == (U # << >> => U[1] # Last(S))
                      => unstutter[S \o U] = unstutter[S] \o unstutter[U]
  <2>1. P(<< >>)
    OBVIOUS
  <2>2. ASSUME NEW U \in Seq(State), P(U), NEW s \in State
        PROVE  P(Append(U,s))
    <3>. DEFINE V == S \o Append(U,s)
    <3>. /\ V \in Seq(State)
         /\ S \o U \in Seq(State)
      OBVIOUS
    <3>. SUFFICES ASSUME Append(U,s)[1] # Last(S)
                  PROVE  unstutter[V] = unstutter[S] \o unstutter[Append(U,s)]
      OBVIOUS
    <3>2. CASE U = << >>
      <4>1. V = Append(S,s)
        BY <3>2, AppendIsConcat
      <4>2. unstutter[V] = Append(unstutter[S], s)
        BY <4>1, unstutter_append
      <4>3. unstutter[Append(U,s)] = Append(<<>>, s)
        BY <3>2, unstutter_append, Zenon
      <4>. QED  BY <4>2, <4>3, AppendIsConcat
    <3>3. CASE U # << >>
      <4>1. Append(U,s)[1] = U[1]
        BY <3>3
      <4>2. unstutter[S \o U] = unstutter[S] \o unstutter[U]
        BY <4>1, <2>2
      <4>3. Last(S \o U) = Last(U)
        BY <3>3 DEF Last
      <4>5. unstutter[V]
            = IF Last(U) = s THEN unstutter[S \o U] ELSE Append(unstutter[S \o U], s)
        BY <3>3, V = Append(S \o U, s), <4>3, unstutter_append
      <4>6. unstutter[Append(U,s)]
            = IF Last(U) = s THEN unstutter[U] ELSE Append(unstutter[U], s)
        BY <3>3, unstutter_append
      <4>. QED  BY <4>5, <4>6, <4>2
    <3>. QED  BY <3>2, <3>3
  <2>3. \A U \in Seq(State) : P(U)
    BY <2>1, <2>2, SequencesInductionAppend, Isa
  <2>. QED  BY <1>2, <2>3, Zenon
<1>. QED  BY <1>1, <1>2


(***************************************************************************)
(* FirstDifferent(s,v) denotes the smallest index m such that s[m] is      *)
(* different from v if such an index exists, and Len(s)+1 otherwise.       *)
(***************************************************************************)
FirstDifferent(s,v) ==
  CHOOSE m \in 1 .. Len(s)+1 :
    /\ \A k \in 1 .. m-1 : s[k] = v
    /\ m = Len(s)+1 \/ s[m] # v

LEMMA FirstDifferentProperties ==
  ASSUME NEW S, NEW s \in Seq(S), NEW v \in S
  PROVE  LET m == FirstDifferent(s,v)
         IN  /\ m \in 1 .. Len(s)+1
             /\ \A k \in 1 .. m-1 : s[k] = v
             /\ m = Len(s)+1 \/ s[m] # v
<1>. DEFINE P(m) == /\ m \in 1 .. Len(s)+1
                    /\ \A k \in 1 .. m-1 : s[k] = v
                    /\ m = Len(s)+1 \/ s[m] # v
<1>1. CASE \A k \in 1 .. Len(s) : s[k] = v
  <2>1. P(Len(s)+1)  BY <1>1
  <2>. QED  BY <2>1 DEF FirstDifferent
<1>2. CASE \E k \in 1 .. Len(s) : s[k] # v
  <2>. DEFINE Q(k) == k \in 1 .. Len(s) /\ s[k] # v
  <2>1. \E k \in Nat : Q(k)  BY <1>2
  <2>2. PICK m \in Nat : /\ Q(m)
                         /\ \A i \in 0 .. m-1 : ~Q(i)
    <3>. HIDE DEF Q
    <3>. QED  BY <2>1, SmallestNatural, Isa
  <2>3. P(m)  BY <2>2
  <2>. QED  BY <2>3, Zenon DEF FirstDifferent
<1>. QED  BY <1>1, <1>2

(***************************************************************************)
(* The characteristic predicate of FirstDifferent uniquely defines the     *)
(* value, hence any value satisfying this predicate equals the result.     *)
(***************************************************************************)
LEMMA IsFirstDifferent ==
  ASSUME NEW S, NEW s \in Seq(S), NEW v \in S,
         NEW m \in 1 .. Len(s)+1,
         \A k \in 1 .. m-1 : s[k] = v,
         m = Len(s)+1 \/ s[m] # v
  PROVE  m = FirstDifferent(s,v)
<1>. DEFINE mm == FirstDifferent(s,v)
<1>1. /\ mm \in 1 .. Len(s)+1
      /\ \A k \in 1 .. mm-1 : s[k] = v
      /\ mm = Len(s)+1 \/ s[mm] # v
  BY FirstDifferentProperties
<1>2. SUFFICES ASSUME m # mm PROVE FALSE
  OBVIOUS
<1>3. CASE m < mm
  <2>. /\ s[m] # v
       /\ m \in 1 .. mm-1
    BY <1>1, <1>3
  <2>. QED  BY <1>1
<1>4. CASE mm < m
  <2>. /\ s[mm] # v
       /\ mm \in 1 .. m-1
    BY <1>1, <1>4
  <2>. QED  OBVIOUS
<1>. QED  BY <1>1, <1>2, <1>3, <1>4

(***************************************************************************)
(* With the help of the preceding operator, we can reduce unstuttering     *)
(* of a concatenation of two sequences to unstuttered versions of the      *)
(* original sequences.                                                     *)
(***************************************************************************)

LEMMA unstutter_concat ==
  ASSUME NEW s \in Seq(State), NEW t \in Seq(State), s # << >>
  PROVE  LET m == FirstDifferent(t, Last(s))
         IN  unstutter[s \o t] = unstutter[s] \o unstutter[SubSeq(t, m, Len(t))]
<1>. DEFINE m == FirstDifferent(t, Last(s))
            left == SubSeq(t, 1, m-1)
            right == SubSeq(t, m, Len(t))
<1>. /\ Last(s) \in State
  BY DEF Last
<1>. /\ 1 \in 1 .. Len(t)+1
     /\ m-1 \in 1-1 .. Len(t)
     /\ m \in 1 .. Len(t)+1
     /\ Len(t) \in m-1 .. Len(t)
     /\ (m-1)+1 = m
     /\ (m-1) - 1 + 1 = m-1
  BY FirstDifferentProperties, Z3
<1>. /\ left \in Seq(State)
     /\ Len(left) = m-1
     /\ right \in Seq(State)
  BY SubSeqProperties, Zenon
<1>1. t = left \o right
  BY ConcatAdjacentSubSeq, SubSeqFull, Zenon
<1>2. \A i \in 1 .. Len(left) : left[i] = Last(s)
  BY FirstDifferentProperties, SubSeqProperties, Z3
<1>3. unstutter[s \o left] = unstutter[s]
  BY <1>2, unstutter_concat_stuttered_last, Zenon
<1>4. unstutter[(s \o left) \o right] = unstutter[s \o left] \o unstutter[right]
  <2>1. ASSUME right # << >>
        PROVE  right[1] # Last(s \o left)
    <3>1. m # Len(t)+1
      BY <2>1, SubSeqEmpty
    <3>2. right[1] = t[m]
      BY <3>1, SubSeqProperties, Z3
    <3>3. right[1] # Last(s)
      BY <3>1, <3>2, FirstDifferentProperties
    <3>4. Last(s \o left) = Last(s)
      <4>. HIDE DEF left
      <4>. QED  BY <1>2 DEF Last
    <3>. QED  BY <3>3, <3>4
  <2>. QED  BY s \o left \in Seq(State), <2>1, unstutter_concat_different, Zenon
<1>. QED  BY <1>1, <1>3, <1>4, ConcatAssociative, Isa

(***************************************************************************)
(* It follows that unstuttering preserves prefixes.                        *)
(***************************************************************************)
LEMMA unstutter_IsPrefix ==
  ASSUME NEW s \in Seq(State), NEW t \in Seq(State), IsPrefix(s,t)
  PROVE  IsPrefix(unstutter[s], unstutter[t])
<1>1. CASE s = << >>
  BY <1>1, unstutter_empty, unstutter_type, EmptyIsPrefix, Zenon
<1>2. CASE s # << >>
  <2>1. PICK u \in Seq(State) : t = s \o u
    BY IsPrefixProperties
  <2>. DEFINE m == FirstDifferent(u, Last(s))
  <2>. Last(s) \in State
    BY <1>2 DEF Last
  <2>2. unstutter[t] = unstutter[s] \o unstutter[SubSeq(u, m, Len(u))]
    BY <1>2, <2>1, unstutter_concat, Zenon
  <2>3. /\ m \in 1 .. Len(u)+1
        /\ Len(u) \in m-1 .. Len(u)
    BY FirstDifferentProperties, Z3
  <2>4. SubSeq(u, m, Len(u)) \in Seq(State)
    BY <2>3, SubSeqProperties
  <2>5. /\ unstutter[s] \in Seq(State)
        /\ unstutter[t] \in Seq(State)
        /\ unstutter[SubSeq(u, m, Len(u))] \in Seq(State)
    BY <2>4, unstutter_type, Zenon
  <2>. QED  BY <2>2, <2>5, IsPrefixProperties, Zenon
<1>. QED  BY <1>1, <1>2

(***************************************************************************)
(* To every prefix of the unstuttered sequence corresponds a prefix of     *)
(* the original sequence.                                                  *)
(* This theorem could be strengthened to assert the existence of a         *)
(* strictly monotonic function that to every prefix l of the unstuttered   *)
(* sequence assigns a prefix r of the original sequence such that          *)
(* l = unstutter[r].                                                       *)
(***************************************************************************)
LEMMA unstutter_correspondence ==
  ASSUME NEW s \in Seq(State), NEW n \in 0 .. Len(unstutter[s])
  PROVE  \E m \in n .. Len(s) : unstutter[SubSeq(s,1,m)] = SubSeq(unstutter[s],1,n)
<1>. DEFINE P(k) == k \leq Len(unstutter[s]) =>
                    \E m \in k .. Len(s) : unstutter[SubSeq(s,1,m)] = SubSeq(unstutter[s],1,k)
<1>. unstutter[s] \in Seq(State)
  BY unstutter_type
<1>1. P(0)
  <2>1. /\ SubSeq(unstutter[s], 1, 0) = << >>
        /\ SubSeq(s, 1, 0) = << >>
    BY SubSeqEmpty
  <2>2. /\ 0 \in 0 .. Len(s)
        /\ SubSeq(s, 1, 0) \in Seq(State)
    BY SubSeqProperties
  <2>. QED  BY <2>1, <2>2, unstutter_empty
<1>2. ASSUME NEW k \in Nat, P(k)
      PROVE  P(k+1)
  <2>1. HAVE k+1 \leq Len(unstutter[s])
  <2>2. s # << >>
    BY <2>1, unstutter_empty
  <2>3. CASE k = 0
    <3>2. /\ 1 \in 1 .. Len(s)
          /\ SubSeq(s, 1, 1) \in Seq(State)
          /\ SubSeq(s, 1, 1) # << >>
          /\ Len(SubSeq(s, 1, 1)) <= 1
      BY <2>2, SubSeqProperties, SubSeqEmpty
    <3>3. unstutter[SubSeq(s,1,1)] = SubSeq(s,1,1)
      BY <3>2, unstutter_type, Zenon DEF unstutter_body
    <3>4. unstutter[s][1] = s[1]
      BY <2>2, unstutter_first
    <3>5. SubSeq(unstutter[s], 1, 1) = SubSeq(s,1,1)
      <4>. /\ SubSeq(unstutter[s], 1, 1) \in Seq(State)
           /\ SubSeq(s, 1, 1) \in Seq(State)
           /\ Len(SubSeq(unstutter[s], 1, 1)) = 1
           /\ Len(SubSeq(s, 1, 1)) = 1
           /\ \A i \in 1 .. 1 : SubSeq(unstutter[s], 1, 1)[i] = SubSeq(s, 1, 1)[i]
        BY <2>1, <2>2, <2>3, <3>4, SubSeqProperties
      <4>. QED  BY SeqEqual
    <3>. QED  BY <2>3, <3>2, <3>3, <3>5, Isa
  <2>4. CASE k > 0
    <3>1. PICK m \in k .. Len(s) : unstutter[SubSeq(s,1,m)] = SubSeq(unstutter[s],1,k)
      BY <1>2, <2>1
    <3>. DEFINE left == SubSeq(s,1,m)
                right == SubSeq(s, m+1, Len(s))
                mm == FirstDifferent(right, s[m])
    <3>. /\ left \in Seq(State)
         /\ Len(left) = m
         /\ left # << >>
         /\ left[m] = s[m]
         /\ s[m] \in State
         /\ SubSeq(unstutter[s],1,k) \in Seq(State)
         /\ Len(SubSeq(unstutter[s],1,k)) = k
         /\ Len(unstutter[left]) = k
         /\ right \in Seq(State)
         /\ Len(right) = Len(s) - m
      <4>1. /\ left \in Seq(State)
            /\ Len(left) = m
        BY 1 \in 1 .. Len(s)+1, m \in 1-1 .. Len(s), m-1+1 = m, SubSeqProperties, Zenon
      <4>2. /\ m \in 1 .. Len(left)
            /\ s[m] \in State
        BY <4>1, <2>4
      <4>3. /\ left # << >>
            /\ left[m] = s[m]
        BY <4>2, SubSeqProperties
      <4>4. /\ SubSeq(unstutter[s],1,k) \in Seq(State)
            /\ Len(SubSeq(unstutter[s],1,k)) = k
        BY <2>1, 1 \in 1 .. Len(unstutter[s])+1, k \in 1-1 .. Len(unstutter[s]), 
           k-1+1 = k, SubSeqProperties, Zenon
      <4>5. Len(unstutter[left]) = k
        BY <3>1, <4>4
      <4>6. /\ right \in Seq(State)
            /\ Len(right) = Len(s) - m
        BY m+1 \in 1 .. Len(s)+1, Len(s) \in (m+1)-1 .. Len(s), 
           Len(s) - (m+1) + 1 = Len(s) - m, SubSeqProperties, Zenon
      <4>. QED  BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6
    <3>. unstutter[left] \in Seq(State)
      BY unstutter_type
    <3>2. /\ mm \in 1 .. Len(right)+1
          /\ \A i \in 1 .. mm-1 : right[i] = s[m]
          /\ mm = Len(right)+1 \/ right[mm] # s[m]
      <4>. HIDE DEF right
      <4>. QED  BY FirstDifferentProperties
    <3>3. s = left \o right
      <4>. /\ 1 \in 1 .. Len(s)+1
           /\ m \in 1-1 .. Len(s)
           /\ Len(s) \in m .. Len(s)
        OBVIOUS
      <4>. QED  BY ConcatAdjacentSubSeq, SubSeqFull, Zenon
    <3>. HIDE DEF mm
    <3>4. mm # Len(right)+1
      <4>. HIDE DEF left, right
      <4>. SUFFICES ASSUME mm = Len(right)+1 PROVE FALSE
        OBVIOUS
      <4>1. \A i \in 1 .. Len(right) : right[i] = s[m]
        BY <3>2
      <4>2. unstutter[s] = unstutter[left]
        BY <3>3, <4>1, unstutter_concat_stuttered_last DEF Last
      <4>. QED  BY <4>2, <2>1
    <3>5. m+mm \in k+1 .. Len(s)
      BY <3>2, <3>4
    <3>. s[m+mm] \in State
      BY <3>5
    <3>6. SubSeq(unstutter[s], 1, k+1) = Append(SubSeq(unstutter[s],1,k), s[m+mm])
      <4>. HIDE DEF left, right
      <4>1. unstutter[s] = unstutter[left] \o unstutter[SubSeq(right, mm, Len(right))]
        BY <3>3, unstutter_concat, Isa DEF Last, mm
      <4>. /\ SubSeq(right, mm, Len(right)) \in Seq(State)
           /\ Len(SubSeq(right, mm, Len(right))) = Len(right) - mm + 1
        BY <3>2, <3>4, SubSeqProperties
      <4>. unstutter[SubSeq(right, mm, Len(right))] \in Seq(State)
        BY unstutter_type, Zenon
      <4>. k+1 \in 1 .. Len(unstutter[s])
        BY <2>1
      <4>2. unstutter[s][k+1] = unstutter[SubSeq(right, mm, Len(right))][1]
        BY <4>1, ConcatProperties
      <4>3. SubSeq(right, mm, Len(right)) # << >>
        BY <3>2, <3>4, EmptySeq, Z3
      <4>4. unstutter[SubSeq(right, mm, Len(right))][1] = SubSeq(right, mm, Len(right))[1]
        BY <4>3, unstutter_first, Zenon
      <4>5. SubSeq(right, mm, Len(right))[1] = s[m+mm]
        <5>1. SubSeq(right, mm, Len(right))[1] = right[mm]
          BY <3>2, <3>4, SubSeqProperties
        <5>2. /\ m+1 \in 1 .. Len(s)+1
              /\ Len(s) \in (m+1)-1 .. Len(s)
              /\ mm \in 1 .. Len(s) - (m+1) + 1
          BY <3>1, <3>2, <3>4
        <5>3. right[mm] = s[m+mm]
          BY <5>2, SubSeqProperties, Z3 DEF right
        <5>. QED  BY <5>1, <5>3
      <4>6. unstutter[s][k+1] = s[m+mm]
        BY <4>2, <4>4, <4>5
      <4>8. SubSeq(unstutter[s], 1, k+1) = SubSeq(unstutter[s], 1, k) \o << unstutter[s][k+1] >>
        BY 1 \in 1 .. Len(unstutter[s]), k+1 \in k .. Len(unstutter[s]), SubSeqRecursiveSecond, Isa
      <4>. QED  BY <4>6, <4>8, AppendIsConcat, Zenon
    <3>. DEFINE mid == SubSeq(right, 1, mm)
                witness == SubSeq(s, 1, m+mm)
    <3>a. mid = SubSeq(s, m+1, m+mm)
      BY <3>2, <3>4, mySubSeqOfSubSeq
    <3>. /\ mid \in Seq(State)
         /\ Len(mid) = mm
         /\ mid[mm] = s[m+mm]
         /\ witness \in Seq(State)
         /\ Len(witness) = m + mm
      BY <3>a, <3>2, <3>5, SubSeqProperties
    <3>10. witness = left \o mid
      <4>. /\ 1 \in 1 .. Len(s)+1
           /\ m \in 1-1 .. Len(s)
           /\ m+mm \in m .. Len(s)
        BY <3>2, <3>5
      <4>. QED  BY <3>a, ConcatAdjacentSubSeq, Zenon
    <3>. HIDE DEF right
    <3>11. FirstDifferent(mid, s[m]) = mm
      <4>. \A i \in 1 .. mm : mid[i] = right[i]
        BY <3>2, <3>4, SubSeqProperties
      <4> mid \in Seq(State)
        OBVIOUS
      <4> HIDE DEF mid
      <4>. QED  BY <3>2, <3>4, IsFirstDifferent, Z3
    <3>12. unstutter[witness] = Append(unstutter[left], s[m+mm])
      <4>1. unstutter[witness] = unstutter[left] \o unstutter[SubSeq(mid, mm, mm)]
        BY unstutter_concat, <3>10, <3>11, Isa DEF Last
      <4>2. /\ SubSeq(mid, mm, mm) \in Seq(State)
            /\ SubSeq(mid, mm, mm) = << s[m+mm] >>
            /\ << s[m+mm] >> # << >>
            /\ Len(<< s[m+mm] >>) \leq 1
        BY SubSeqProperties
      <4>3. unstutter[SubSeq(mid, mm, mm)] = << s[m+mm] >>
        BY <4>2, unstutter_type, Zenon DEF unstutter_body
      <4>. QED  BY <4>1, <4>3, AppendIsConcat, Zenon
    <3>. WITNESS m+mm \in k+1 .. Len(s)
    <3>. QED  BY <3>1, <3>6, <3>12, Zenon
  <2>. QED  BY <2>3, <2>4
<1>3. \A k \in Nat : P(k)
  BY <1>1, <1>2, NatInduction, Isa
<1>. QED  BY <1>3

(***************************************************************************)
(* The following lemma is a weak reciprocal of unstutter_IsPrefix:         *)
(* If the unstuttered version of sequence s is a prefix of the unstuttered *)
(* version of sequence t then there is a prefix of t that is stuttering    *)
(* equivalent to s.                                                        *)
(***************************************************************************)
LEMMA IsPrefix_unstutter ==
  ASSUME NEW s \in Seq(State), NEW t \in Seq(State),
         IsPrefix(unstutter[s], unstutter[t])
  PROVE  \E u \in Seq(State) : IsPrefix(u,t) /\ unstutter[u] = unstutter[s]
<1>. /\ unstutter[s] \in Seq(State)
     /\ unstutter[t] \in Seq(State)
  BY unstutter_type
<1>1. /\ Len(unstutter[s]) \in 0 .. Len(unstutter[t])
      /\ unstutter[s] = SubSeq(unstutter[t], 1, Len(unstutter[s]))
  BY IsPrefixProperties
<1>2. PICK m \in Len(unstutter[s]) .. Len(t) : unstutter[SubSeq(t,1,m)] = unstutter[s]
  BY <1>1, unstutter_correspondence, Zenon
<1>. /\ 1 \in 1 .. Len(t)+1
     /\ m \in 1-1 .. Len(t)
     /\ m-1+1 = m
  OBVIOUS
<1>3. /\ SubSeq(t,1,m) \in Seq(State)
      /\ Len(SubSeq(t,1,m)) = m
  BY SubSeqProperties, Zenon
<1>4. IsPrefix(SubSeq(t,1,m), t)
  BY <1>3, m <= Len(t), IsPrefixProperties, Zenon
<1>. QED  BY <1>2, <1>3, <1>4

(***************************************************************************)
(* filter(H,a) projects the finite sequence of system states H to the      *)
(* states of agent a and then removes stuttering.                          *)
(***************************************************************************)
filter(H, a) == unstutter[project_agent(H,a)]

LEMMA filter_type == 
  ASSUME NEW H \in Seq(SystemState), NEW a \in Agent
  PROVE  filter(H,a) \in Seq(State)
BY unstutter_type, project_agent_type DEF filter

LEMMA filter_empty ==
  ASSUME NEW H \in Seq(SystemState), NEW a \in Agent
  PROVE  filter(H,a) = <<>> <=> H = <<>>
BY project_agent_type, unstutter_empty DEF filter

LEMMA filter_range ==
  ASSUME NEW H \in Seq(SystemState), NEW a \in Agent
  PROVE  Range(filter(H,a)) = { H[i].st[a] : i \in 1 .. Len(H) }
<1>1. project_agent(H,a) \in Seq(State)
  BY project_agent_type
<1>2. Range(filter(H,a)) = Range(project_agent(H,a))
  BY <1>1, unstutter_range DEF filter
<1>. QED  BY <1>1, <1>2, LenProperties, Zenon DEF Range, project_agent

LEMMA Last_filter ==
  ASSUME NEW H \in Seq(SystemState), H # << >>, NEW a \in Agent
  PROVE Last(filter(H,a)) = Last(H).st[a]
<1>. /\ project_agent(H,a) \in Seq(State)
     /\ project_agent(H,a) # << >>
  BY project_agent_type
<1>. QED  BY unstutter_last DEF filter, project_agent, Last

LEMMA filter_stutter ==
  ASSUME NEW H \in Seq(SystemState), H # <<>>,
         NEW a \in Agent,
         NEW syst \in SystemState, syst.st[a] = Last(H).st[a]
  PROVE  filter(Append(H, syst), a) = filter(H, a)
<1>. /\ project_agent(H,a) \in Seq(State)
     /\ project_agent(H,a) # << >>
  BY project_agent_type
<1>. project_agent(Append(H,syst), a)
     = Append(project_agent(H,a), Last(project_agent(H,a)))
  BY project_agent_append, project_agent_last
<1>. QED  BY unstutter_append DEF filter, Last

LEMMA filter_append ==
  ASSUME NEW H1 \in Seq(SystemState), NEW H2 \in Seq(SystemState),
         NEW syst1 \in SystemState, NEW syst2 \in SystemState, NEW a \in Agent,
         filter(H1, a) = filter(H2, a), syst1.st[a] = syst2.st[a]
  PROVE  filter(Append(H1, syst1), a) = filter(Append(H2, syst2), a)
<1>1. CASE H1 = << >>
  <2>1. H2 = << >>
    BY <1>1, filter_empty, Zenon
  <2>2. project_agent(Append(H1,syst1), a) = project_agent(Append(H2,syst2), a)
    BY <1>1, <2>1, project_agent_append, Isa
  <2>. QED  BY <2>2 DEF filter
<1>2. CASE H1 # << >>
  <2>. /\ project_agent(H1,a) \in Seq(State)
       /\ project_agent(H2,a) \in Seq(State)
       /\ syst1.st[a] \in State
       /\ project_agent(H1,a) # << >>
       /\ project_agent(H2,a) # << >>
    BY <1>2, project_agent_type, filter_empty DEF SystemState
  <2>. Last(project_agent(H1,a)) = Last(project_agent(H2,a))
    BY unstutter_last DEF filter
  <2>. QED  BY project_agent_append, unstutter_append, Isa DEF filter
<1>. QED  BY <1>1, <1>2

LEMMA filter_IsPrefix ==
  ASSUME NEW H1 \in Seq(SystemState), NEW H2 \in Seq(SystemState),
         NEW a \in Agent,
         IsPrefix(H1, H2)
  PROVE  IsPrefix(filter(H1,a), filter(H2,a))
<1>. /\ project_agent(H1,a) \in Seq(State)
     /\ project_agent(H2,a) \in Seq(State)
     /\ Len(project_agent(H1,a)) = Len(H1)
     /\ Len(project_agent(H2,a)) = Len(H2)
  BY project_agent_type
<1>. IsPrefix(project_agent(H1,a), project_agent(H2,a))
  <2>. /\ Len(H1) <= Len(H2)
       /\ H1 = Restrict(H2, 1 .. Len(H1))
    BY IsPrefixProperties, LenProperties
  <2>. project_agent(H1,a) = Restrict(project_agent(H2,a), 1 .. Len(H1))
    BY DEF project_agent, Restrict
  <2>. QED  BY IsPrefixProperties, LenProperties
<1>. QED  BY unstutter_IsPrefix DEF filter

LEMMA IsPrefix_filter ==
  ASSUME NEW H1 \in Seq(SystemState), NEW H2 \in Seq(SystemState), NEW a \in Agent,
         IsPrefix(filter(H1,a), filter(H2,a))
  PROVE  \E H \in Seq(SystemState) : IsPrefix(H, H2) /\ filter(H,a) = filter(H1,a)
<1>. /\ project_agent(H1,a) \in Seq(State)
     /\ project_agent(H2,a) \in Seq(State)
  BY project_agent_type
<1>1. PICK Ha \in Seq(State) :
           /\ IsPrefix(Ha, project_agent(H2,a))
           /\ unstutter[Ha] = filter(H1,a)
  BY IsPrefix_unstutter DEF filter
<1>. DEFINE H == [ i \in 1 .. Len(Ha) |-> H2[i] ]
<1>. /\ Len(Ha) <= Len(H2)
     /\ Ha = Restrict(project_agent(H2,a), 1 .. Len(Ha))
     /\ H \in Seq(SystemState)
     /\ Len(H) = Len(Ha)
  BY <1>1, IsPrefixProperties, project_agent_type, LenProperties 
<1>. \A i \in 1 .. Len(Ha) : Ha[i] = H2[i].st[a]
  BY DEF Restrict, project_agent
<1>2. project_agent(H, a) = Ha
  <2>. /\ project_agent(H,a) \in Seq(State)
       /\ Len(project_agent(H,a)) = Len(H)
    BY project_agent_type, Zenon
  <2>. \A i \in 1 .. Len(Ha) : project_agent(H,a)[i] = Ha[i]
    BY DEF project_agent
  <2>. QED  BY SeqEqual
<1>3. filter(H,a) = filter(H1,a)
  BY <1>1, <1>2 DEF filter
<1>4. IsPrefix(H, H2)
  BY IsPrefixProperties, LenProperties, Zenon DEF Restrict
<1>. QED  BY <1>3, <1>4, Zenon

=============================================================================
\* Modification History
\* Last modified Mon Jan 25 16:57:47 CET 2016 by merz
\* Last modified Wed Aug 26 15:33:14 CEST 2015 by matthieu
\* Last modified Wed Jul 15 15:35:36 CEST 2015 by merz
